/*	16F877 BOOT LOADER Copyright (C)1999 HI-TECH Software.
 *	Freely distubutable.
 *
 *	Linker options should be changed to position the bootloader at
 *	at the desired address. ie: 0x1F00 - 0x1FFF
 *	Designed to be < 256 words. Compile with full 
 *	optimizations (-Zg -O) to achieve this.
 *
 */

#include <pic.h>

#include "bootldr.h"
#include "proto.h"


/*
 * Set LoaderStart to the address where the boot loader will start,
 * the other #defines will adjust accordingly
 *
 */

#define LoaderStart 0x1F00

/*
 * Addresses 0-4 of the downloaded program must be relocated so
 * that they don't overwrite this bootloader startup. UserStart is the
 * address where they will be relocated.
 *
 * LoaderLOW and LoaderHI are the low and high addresses where the
 * bootloader code lives. This is needed so that a downloaded program
 * can't overwrite the bootloader.
 *
 * LoaderID is the address where IDValue is written when a program
 * is downloaded. This is used to determine if a program has been
 * downloaded yet.
 *
 */

#define UserStart	LoaderStart		  /* user addresses 0-4 will go here */
#define LoaderSize	0xFF			  /* compiled size of loader (words) */	
#define LoaderLOW	(LoaderStart+5)	          /* Low address of boot loader */
#define LoaderHI	(LoaderStart+LoaderSize)  /* High address of boot loader */
#define LoaderID	(LoaderStart+4)           /* Location to check if program is installed */
#define IDValue		0xAA      /* byte to write/check at LoaderID address */

/*
 * MaxBuf is the maximum buffer size allocated for receiving the
 * incoming data.
 *
 * MaxRetry is the maximum number of attempts made at writing the
 * data.
 *
 */

#define MaxBuf 		32        /* Maximum size of buffer */
#define MaxRetry	2	  /* Maximum number of write retries */	


/*
 * Comms setup:
 */
 
#define BAUD 9600      
#define FOSC 4000000L
#define NINE_BITS 0     /* Use 9bit communication? FALSE=8bit */
#define DIVIDER ((int)(FOSC/(16UL * BAUD) -1))
#define HIGH_SPEED 1

/* if DIVIDER is >255 use: 
#define DIVIDER ((int)(FOSC/(64UL * BAUD) -1)
#define HIGH_SPEED 0
*/

/*********************************************************************/

unsigned char buff[MaxBuf];	     
unsigned char count,amount,chk1,chk2,retry;
unsigned address,tmpaddr;


/* Serial initialization */
#define InitComms() 		\
 	BRGH = HIGH_SPEED;	\
        SPBRG = DIVIDER;     	\
	SYNC = 0;		\
	SPEN = 1;		\
	CREN = 1;		\
	SREN = 0;		\
	TXIE = 0;		\
	RCIE = 0;		\
        TX9  = NINE_BITS;	\
        RX9  = NINE_BITS;	\
  	TXEN = 1

void 
putbyte(unsigned char byte) {
	/* output one byte */
	while(!TXIF)	/* set when register is empty */
		continue;
	TXREG = byte;
}

unsigned char 
getbyte() {
	/* retrieve one byte */
	while(!RCIF)	/* set when register is not empty */
		continue;
	return RCREG;	/* RXD9 and FERR are gone now */
}

unsigned char 
WriteData(void) {
	/* write to program memory */
	
	tmpaddr = address;
	retry=MaxRetry;
	count=0;
	while (count<amount) {
	        /* make sure it doesn't overwrite the loader */
              if ((tmpaddr >= LoaderLOW) && (tmpaddr <= LoaderHI)) 
	        	goto next_addr; /* ignore, do next address */
	        	
	      /* move user's start code to a different location */
	      EEADR=tmpaddr;  
              if (tmpaddr < 4) 
                        EEADRH=UserStart >> 8;
              else
	      		EEADRH=tmpaddr >> 8;
              EEDATH = buff[count];
              EEDATA = buff[count+1];
              EEPGD = 1;
              WREN=1;
              EECON2 = 0x55;	/* write sequence... */
              EECON2 = 0xaa;
              WR=1;		
              asm("nop");
              asm("nop");	/* wait for write to complete */
              WREN=0;
              EEPGD=1;
              RD=1;		/* read back to verify */
              asm("nop");	/* wait for read */
              asm("nop");
              if ((EEDATH != buff[count]) || (EEDATA != buff[count+1])) {  
		        if (--retry) 
			        continue;	/* bad write, try again */
                        else 
				return 0;       /* still didn't work, exit.. */ 
                }
next_addr:
		count+=2;	/* write successful, do next address */
		++tmpaddr;
	}
	return 1;	/* successful write of buffer, return true */
}


void 
main(void) {

	/* If DOWNLOAD is false and there has been a user program
	   downloaded - it will run the user program. Otherwise it will 
	   wait until DOWNLOAD is true. By default, DOWNLOAD is set to
	   pin RB0.
	   The check to see if a program has be downloaded doesn't have to
	   be in the for loop. It is done this way to reduce code size.
	*/
		   
	for(DOWNPORT=INPUT; !DOWNLOAD;) {	/* wait for DOWNLOAD */
		EEADR=LoaderID & 0xFF;		/* read LoaderID address */
		EEADRH=LoaderID >> 8;
		EEPGD=1;
		RD=1;
		asm("nop");			/* wait for read */
		asm("nop");
		if (EEDATA == IDValue) {	/* check if a program is installed */
		#asm
   		    ljmp UserStart	/* run the downloaded program */
		#endasm
		}
	}
	InitComms(); 		
	/* main loader loop */
	for(;;) 
	{
		switch (getbyte()) 
		{
		case IDENT:			/* identify loader */
                        putbyte(IDACK);		/* acknowledge */
			continue;
		case WRITE:			/* write & verify data */
			address = getbyte() << 8;	/* get write address */
			address |= getbyte();
			amount = getbyte();     /* assume amount < MaxBuf */
			chk2 = getbyte();       /* get checksum */
			count=chk1=0;
			while (count < amount) {	/* receive data */
				buff[count] = getbyte();
				chk1+=buff[count++];
			}
			if (chk1 != chk2) {  /* verify check sum */
			    putbyte(DATA_BAD);
			    continue;
			}
			putbyte(DATA_OK);
                        putbyte(WriteData()?WOK:WBAD);  /* write data */
			continue;
		case DONE:			/* finish downloading */
			address=LoaderID;	/* write LoaderID */
			amount=2;
			buff[0] = 0;
			buff[1] = IDValue;
                        putbyte(WriteData()?WOK:WBAD);
		}
	}			
}

